#  Copyright (C) 2003-2005 Joe Mallon <jmmallon at joescafe dot com>
#  Reads data from iTunesDB & generates list of songs and playlists on iPod
#  URL: http://www.joescafe.com/tunes2html
#
#  Uses code from tunes2pod.pl in GNUPod
#  Copyright (C) 2002-2003 Adrian Ulrich <pab at blinkenlights.ch>
#  Part of the gnupod-tools collection
#
#  URL: http://www.gnu.org/software/gnupod/
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details:
#           http://www.gnu.org/licenses/gpl.txt
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#  This script has been tested on Windows 98, Windows 2000 Professional,
#  Windows ME, Windows XP, Linux, Mac OS X, Solaris 5.7, and Solaris 5.8, using
#  Perl 5.005 and Perl 6.
#
#  iTunes and iPod are trademarks of Apple
#
#  This product is not supported/written/published by Apple!

# header of a valid iTunesDB
my $ipodmagic = "6d 68 62 64 68 00 00 00";
my $config_file = "tunes2html.cfg";
$| = "1";
$ver = "4.0.2";
$print = 1;

if ($ARGV[0] eq '-s') {
	$print = 0;
	shift(@ARGV);
}

if ($print) {
	print "\ntunes2html $ver (C) 2003-2005 Joe Mallon\n";
	print "This tool reads data from an iTunesDB file & generates a\n";
	print "list of songs and playlists on the iPod\n\n";
	print "For the latest version, check out http://www.joescafe.com/tunes2html\n\n";
	print "Uses code from tunes2pod.pl 0.6 (C) 2002-2003 Adrian Ulrich\n";
	print "This tool is covered by the GNU General Public License\n\n";
}
exit 0 if ($ARGV[0] eq '-v');

my %config =  (
	'iTunesDB' => './iTunesDB',
	'ArtistColor' => '#000070',
	'AlbumColor' => '#007000',
	'SongColor' => '#007000',
	'PlaylistColor' => '#007000',
	'TextColor' => 'black',
	'BGColor' => 'white',
	'StatsColor' => 'ivory',
	'UseComposer' => 'No',
	'TimeColor' => 'lightgrey',
	'Title' => 'iPod Contents',
	'OutputFormat' => 'Both',
	'AddLinks' => 'No',
	'ShowDots' => 'Yes',
	'AlbumSort' => 'chron',
	'SongSort' => 'track',
	'VarTrackSort' => 'song',
	'PlusImage' => 'plusbl.gif',
	'MinusImage' => 'minusbl.gif',
	'FontFamily' => 'sans-serif',
	'HTMLFilename' => 'iPod.htm',
	'TextFilename' => 'iPod.txt',
	'DupesFilename' => 'dupes.txt',
	'DateFilename' => 'No'
);

my %artists = my %playlists = my %collection = ();

print "Reading configuration.\n" if ($print);
my $configvar = read_config_file($config_file,\%config);
%config = %$configvar;
if (@ARGV) {
	print "Processing command line.\n" if ($print);
	$configvar = read_command_line(\@ARGV,\%config);
	%config = %$configvar;
}

my $output;
if (lc($config{'OutputFormat'}) eq 'text') {
	$output = 't';
} elsif (lc($config{'OutputFormat'}) eq 'html') {
	$output = 'h';
} else {
	$output = 'b';
}

$config{'AlbumSort'} = (lc($config{'AlbumSort'}) eq 'alpha') ? 'alpha' : 'chron';
$config{'SongSort'} = (lc($config{'SongSort'}) eq 'alpha') ? 'alpha' : 'track';
$config{'VarTrackSort'} = (lc($config{'VarTrackSort'}) eq 'artist') ? 'artist' : 'song';

print "\nCollecting song data.\n" if ($print);
my ($infoptr,$playlist,$tracks,$sec,$size,$dupes) = parsomatic($config{'iTunesDB'});
%collection = %$infoptr;
%playlists = %$playlist;
@dupes = @$dupes;

my($al, $allen, $alstr, $alstrx, $altr, $arlen, $arstrx, $arstr, $art, $artr, $cd, $cds, $count, $sets, $snum, $ti, $time, $tr, $var, $yr, @sin, @songs, @var);

foreach $al (sort byart (keys(%collection))) {
 	$art = (defined($config{'UseComposer'}) and lc($config{'UseComposer'}) eq 'yes' and $collection{$al}->{'sortart'}) ? $collection{$al}->{'sortart'} : $collection{$al}->{'artist'};
	$alcomp = $al;
	$alcomp =~ s/(\W)/\\$1/g;
	push(@{$artists{$art}->{'albums'}},$al) unless (grep(/^$alcomp$/,@{$artists{$art}->{'albums'}}));
}

$sets = keys(%artists);
$cds = keys(%collection);
$size = sprintf("%.01f MB",$size/(1024*1024)) . " - " . sprintf("%.01f GB",$size/(1024*1024*1024));
$ti = int($sec / 86400) . "d ";
$sec = $sec % 86400;
$ti .= int($sec / 3600) . "h ";
$sec = $sec % 3600;
$ti .= int($sec/60) . "m " . $sec % 60 . "s";
my ($s,$m,$h,$d,$o,$y) = (localtime)[0,1,2,3,4,5];
$time = ($y + 1900) . "/" . sprintf("%02d",$o + 1) . "/" . sprintf("%02d",$d) . "  " . sprintf("%02d",$h) . ":" . sprintf("%02d",$m) . ":" . sprintf("%02d",$s);
my $filetime = ($y + 1900) . sprintf("%02d",$o + 1) . sprintf("%02d",$d);
if (lc($config{'DateFilename'}) eq 'yes') {
	$config{'HTMLFilename'} =~ s/\.([^\.]+)$/-$filetime.$1/;
	$config{'TextFilename'} =~ s/\.([^\.]+)$/-$filetime.$1/;
}

my @outphrase = ();
my $no_open = 0;
if ($output eq 'h' or $output eq 'b') {
	open(OUTHTML, ">$config{'HTMLFilename'}") or $no_open = 1;
	if ($no_open) {
		print "Could not write to $config{'HTMLFilename'}: $!\nUsing iPod.htm\n";
		$config{'HTMLFilename'} = "./iPod.htm";
		open(OUTHTML, ">$config{'HTMLFilename'}") or die("Could not write to $config{'HTMLFilename'}: $!\n");
	}
	push(@outphrase,$config{'HTMLFilename'});
}

$no_open = 0;
if ($output eq 't' or $output eq 'b') {
	open(OUTTXT, ">$config{'TextFilename'}") or $no_open = 1;
	if ($no_open) {
		print "Could not write to $config{'TextFilename'}: $!\nUsing iPod.txt\n";
		$config{'TextFilename'} = "./iPod.txt";
		open(OUTTXT, ">$config{'TextFilename'}") or die("Could not write to $config{'TextFilename'}: $!\n");
	}
	push(@outphrase,$config{'TextFilename'});
}

print "\nCreating " . join(' and ',@outphrase) . ".\n" if ($print);

if ($output eq 't' or $output eq 'b') {
	print OUTTXT << "ENDOUTTXT";
$config{'Title'} - $time

This page generated by tunes2html

Statistics
Sets: $sets
Discs: $cds
Tracks: $tracks
Playing Time: $ti
Space Used: $size

SINGLE ARTISTS

ENDOUTTXT
}

if ($output eq 'h' or $output eq 'b') {
	print OUTHTML << "ENDOUTHTML";
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-US"
lang="en-US">
<HEAD><TITLE>$config{'Title'}</TITLE>
<style>
H4{margin-bottom:5px;color:$config{'TimeColor'}}
IMG{margin-right:3px;margin-left:10px;margin-top 1px}
LI{font-size:10pt;color:$config{'SongColor'};margin:1px}
OL{margin-top:0px;margin-bottom:0px}
TD{FONT-FAMILY:$config{'FontFamily'};FONT-SIZE:9pt;}
BODY{margin:0px;BACKGROUND-COLOR:$config{'BGColor'};COLOR:$config{'TextColor'};FONT-FAMILY:$config{'FontFamily'};FONT-SIZE:8pt}
.tdMain1{border:1px solid tan;background:$config{'BGColor'}}
.tdMain2{border:1px solid tan;background:$config{'StatsColor'}}
.tdStatHdr{font-weight:bold;color:red}
.bHdr{border-bottom:1px solid tan}
.bArtHdr{font-size:10pt;color:$config{'ArtistColor'};text-decoration: none;cursor:hand}
.bArtHdr:hover{font-size:10pt;color:"red";text-decoration: none;cursor:hand}
.bVAHdr{font-size:10pt;color:$config{'AlbumColor'};text-decoration: none;cursor:hand}
.bVAHdr:hover{font-size:10pt;color:"red";text-decoration: none;cursor:hand}
.bPlaylistHdr{font-size:10pt;color:$config{'PlaylistColor'};text-decoration: none;cursor:hand}
.bPlaylistHdr:hover{font-size:10pt;color:"red";text-decoration: none;cursor:hand}
.bAlbumHdr{font-size:10pt;text-decoration: none;color:$config{'AlbumColor'};cursor:hand;position:relative;left:+15px}
.bAlbumHdr:hover{font-size:10pt;text-decoration: none;color:"red";cursor:hand}
.subImage{position:relative; left:+15px}
</style>

<SCRIPT LANGUAGE="JavaScript">
var plusImg = new Image();
	plusImg.src = "$config{'PlusImage'}"
var minusImg = new Image();
	minusImg.src = "$config{'MinusImage'}"

function showAlbums() {
	if (document.getElementsByTagName)
		x = document.getElementsByTagName('div');
	else if (document.all)
		x = document.all.tags('div');
	for (var i=0;i<x.length;i++)
	{
		if (x[i].id.indexOf("Set") != -1) {
			x[i].style.display = "";
		}
	}
	if (document.getElementsByTagName)
		x = document.getElementsByTagName('img');
	else if (document.all)
		x = document.all.tags('img');
	for (var i=0;i<x.length;i++)
	{
		if (x[i].id.indexOf("Set") != -1) {
			x[i].src = minusImg.src;
		}
	}
}

function hideAlbums() {
	if (document.getElementsByTagName)
		x = document.getElementsByTagName('div');
	else if (document.all)
		x = document.all.tags('div');
	for (var i=0;i<x.length;i++)
	{
		if ((x[i].id.indexOf("Set") != -1) || (x[i].id.indexOf("Album") != -1)) {
			x[i].style.display = "none";
		}
	}
	if (document.getElementsByTagName)
		x = document.getElementsByTagName('img');
	else if (document.all)
		x = document.all.tags('img');
	for (var i=0;i<x.length;i++)
	{
		if ((x[i].id.indexOf("Set") != -1) || (x[i].id.indexOf("Album") != -1)) {
			x[i].src = plusImg.src;
		}
	}
}

function showLevel( _levelId, _imgId ) {
	var thisLevel = document.getElementById( _levelId );
	var thisImg = document.getElementById( _imgId );
	if ( thisLevel.style.display == "none") {
		thisLevel.style.display = "";
		thisImg.src = minusImg.src;
		}
	else {
		thisLevel.style.display = "none";
		thisImg.src = plusImg.src;
		}
}
</SCRIPT>
</HEAD><BODY>
<div id=dvMain>
<table style="border-bottom:1px solid tan" cellspacing=0 cellpadding=5 width="100%">
<tr bgcolor=$config{'BGColor'}><td valign=middle><h2 style="margin-bottom:5px">$config{'Title'}</h2></td><td align=right valign=middle><h4 style="margin-bottom:5px;color:$config{'TimeColor'}">$time</h4></td></tr>
</table>

<table cellspacing=5 cellpadding=5 width="100%">
<tr><td valign=top class=tdMain2>
<b>This page generated by <a href=http://www.joescafe.com/tunes2html>tunes2html</a></b><p>
<b class=bHdr>Statistics</b><p>
<div id=dvStat style="margin-left:20px">
<table cellspacing=0 cellpadding=2>
<tr><td class=tdStatHdr>Sets:</td><td>$sets</td></tr>
<tr><td class=tdStatHdr>Discs:</td><td>$cds</td></tr>
<tr><td class=tdStatHdr>Tracks:</td><td>$tracks</td></tr>
<tr><td class=tdStatHdr>Playing Time:</td><td>$ti</td></tr>
<tr><td class=tdStatHdr>Space Used:</td><td>$size</td></tr>
</table>
</div>
</td></tr>
<tr><td valign=top class=tdMain1>
<a class=bArtHdr href="javascript:showAlbums();"><b>Show All Albums</b></a><br>
<a class=bArtHdr href="javascript:hideAlbums();"><b>Hide All Albums</b></a><p>
<b class=bHdr>Single Artists</b><p>

ENDOUTHTML
}

foreach $art (sort(keys(%artists))) {
	if ($art =~ /Various/) {
		push(@var,$art); 
	} else {
		push(@sin,$art); 
	}
}

my @arts = sort bythe (@sin);
push(@arts,@var);
foreach $art (@arts) {
	if (!$var and $art =~ /:Various/) {
		$var++;
		print OUTTXT "VARIOUS ARTISTS\n\n" if ($output eq 't' or $output eq 'b');
		print OUTHTML "<p><b class=bHdr>Various Artists</b><p>\n" if ($output eq 'h' or $output eq 'b');
	}
	$artp = $arstr = $arstrx = "";
	$cd = @{$artists{$art}->{'albums'}};
	$artr = $arlen = 0;
	@als = (lc($config{'AlbumSort'}) eq 'alpha') ? sort bythe (@{$artists{$art}->{'albums'}}) : sort byyear (@{$artists{$art}->{'albums'}});
	foreach $al (@als) {
		$artp = $collection{$al}->{'artist'} unless ($artp);
		$count++;
		$alstrx = $alstr = "";
		@songs = ();
		$altr = $allen = $snum = 0;
		foreach $d (@{$collection{$al}->{'tracks'}}) {
			foreach $s (@{$d}) {
				push(@songs, $s) if (defined($s) and defined($s->{'song'}) and $s->{'song'});
			}
		}
		unless ($collection{$al}->{'sort'} eq 'track' and $config{'SongSort'} eq 'track') {
			if ($var and $config{'VarTrackSort'} eq 'artist') {
				@songs = sort byartist (@songs);
			} else {
				@songs = sort bysong (@songs);
			}
		}
		foreach $tr (@songs) {
			if ($tr->{'song'}) {
				$a1 = ($config{'AddLinks'} eq 'Yes') ? "<a href='" . $tr->{'file'} . "'>" : "";
				$a2 = ($config{'AddLinks'} eq 'Yes') ? "</a>" : "";
				if ($var) {
					$alstrx .=  "\t" . ++$snum . ". $tr->{'song'} - $tr->{'artist'}";
					$alstr .=  "<li> $a1$tr->{'song'} - $tr->{'artist'}" . $a2;
				} else {
					$alstrx .=  "\t\t" . ++$snum . ". $tr->{'song'}";
					$alstr .=  "<li> $a1$tr->{'song'}" . $a2;
				}
				if ($tr->{'length'}) {
					$alstr .=  " [" . &timemake($tr->{'length'}) . "]";
					$alstrx .=  " [" . &timemake($tr->{'length'}) . "]";
					$allen += $tr->{'length'};
					$arlen += $tr->{'length'};
				}
				$alstr .=  "\n";
				$alstrx .=  "\n";
				$altr++;
				$artr++;
			}
		}
		$yr = $collection{$al}->{'year'} ? "- $collection{$al}->{'year'} " : "";
		if ($var) {
			$arstrx .= qq|$al $yr(T:$altr) [| . &timemake($allen) . qq|]\n$alstrx\n|;
			$arstr .= qq|<a class=bVAHdr href="javascript:showLevel('Album$count','imgAlbum$count');"><img border=0 id=imgAlbum$count src=plusbl.gif><b>$al $yr(T:$altr) [| . &timemake($allen) . qq|]\n</b></a><br><div id=Album$count style='position:relative;left:+15px;display:none'><ol>$alstr</ol></div>\n|;
		} else {
			$arstrx .= qq|\t$al $yr(T:$altr) [| . &timemake($allen) . qq|]\n$alstrx|;
			$arstr .= qq|<a href="javascript:showLevel('Album$count','imgAlbum$count');"><img border=0 id=imgAlbum$count class=subImage src=plusbl.gif></a><a href="javascript:showLevel('Album$count','imgAlbum$count');" class=bAlbumHdr><b>$al $yr(T:$altr) [| . &timemake($allen) . qq|]\n</b></a><br><div id=Album$count style='position:relative;left:+15px;display:none'><ol>$alstr</ol></div>|;
		}
	}
	$count++;
	print "." if ($count % 7 == 0 and lc($config{'ShowDots'}) eq 'yes' and $print);
	if ($var) {
		$arstr =~ s/:\+15/:\+6/g;
		print OUTHTML &uncode($arstr) if ($output eq 'h' or $output eq 'b');
		print OUTTXT &uncode($arstrx) if ($output eq 't' or $output eq 'b');
	} else {
		print OUTTXT &uncode(qq|$artp (D:$cd, T:$artr) [| . &timemake($arlen) . qq|]\n$arstrx\n|) if ($output eq 't' or $output eq 'b');
		print OUTHTML &uncode(qq|<a class=bArtHdr href="javascript:showLevel('Set$count','imgSet$count');"><img id=imgSet$count border=0 src=plusbl.gif><b>$artp (D:$cd, T:$artr) [| . &timemake($arlen) . qq|]\n</b></a><br><div id=Set$count style='display:none'>$arstr</div>\n|) if ($output eq 'h' or $output eq 'b');
	}
}

print OUTHTML "<p><b class=bHdr>Playlists</b><p>\n" if ($output eq 'h' or $output eq 'b');
print OUTTXT "PLAYLISTS\n\n" if ($output eq 't' or $output eq 'b');

my ($playliststxt,$playlistshtml);

foreach $play (sort(keys(%playlists))) {
	my $playlen = my $snum = 0;
	my $songstxt = my $songshtml = "";
	foreach $tr (@{$playlists{$play}}) {
		$count++;
		$playlen += $tr->{'length'};
		$songstxt .= "\t" . ++$snum . ". $tr->{'song'} - $tr->{'artist'} [" . &timemake($tr->{'length'}) . "]\n";
		$songshtml .= "<li> $tr->{'song'} - $tr->{'artist'} [" . &timemake($tr->{'length'}) . "]\n";
	}
	$playliststxt .= "$play (T:" . scalar(@{$playlists{$play}}) . ") ["  . &timemake($playlen) . qq|]\n$songstxt\n|;
	$playlistshtml .= qq|<a class=bPlaylistHdr href="javascript:showLevel('Album$count','imgAlbum$count');"><img border=0 id=imgAlbum$count src=plusbl.gif><b>$play (T:| . scalar(@{$playlists{$play}}) . qq|) [| . &timemake($playlen) . qq|]\n</b></a><br><div id=Album$count style='position:relative;left:+15px;display:none'><ol>$songshtml</ol></div>\n|;
  	$playliststxt .= "\n";
}

print OUTHTML &uncode($playlistshtml) if ($output eq 'h' or $output eq 'b');
print OUTTXT &uncode($playliststxt) if ($output eq 't' or $output eq 'b');
if ($output eq 'h' or $output eq 'b') {
	print OUTHTML << "ENDOUTHTML";
</td></tr></table>
</body></html>
ENDOUTHTML
}

close(OUTHTML) if ($output eq 'h' or $output eq 'b');
close(OUTTXT) if ($output eq 't' or $output eq 'b');
print "\n\nDone!\n" if ($print);

if (@dupes) {
	print "\nSome of these songs may not be listed because they share an album name and\ntrack number:\n" . uncode(join("\n",@dupes)) . "\n\n" if ($print);
	$config{'DupesFilename'} =~ s/\.([^\.]+)$/-$filetime.$1/ if (lc($config{'DateFilename'}) eq 'yes');
	open(OUTDUPES, ">$config{'DupesFilename'}") or $no_open = 1;
	if ($no_open) {
		print "Could not write to $config{'DupesFilename'}: $!\nUsing dupes.txt\n";
		$config{'DupesFilename'} = "./dupes.txt";
		open(OUTDUPES, ">$config{'DupesFilename'}") or die("Could not write to $config{'DupesFilename'}: $!\n");
	}
	print "Creating $config{'DupesFilename'}.\n\n" if ($print);
	print OUTDUPES "\nSome of these songs may not be listed because they share an album name and\ntrack number:\n" . uncode(join("\n",@dupes)) . "\n";
	close(OUTDUPES);
}

exit(0);


sub timemake {
	my ($sec) = @_;
	my $ti = sprintf("%02d",int($sec/3600)) . ":";
	$sec = $sec % 3600;
 	$ti .= sprintf("%02d",int($sec/60)) . ":" . sprintf("%02d",$sec % 60);
	return($ti);
}

sub bythe {
	my $c = $a;
	my $d = $b;
	$c =~ s/^The //;
	$d =~ s/^The //;
	my $ret = uc($c) cmp uc($d);
	return($ret) if ($ret);
	return(0);
}

sub byyear {
	my $ret = $collection{$a}->{'year'} <=> $collection{$b}->{'year'};
	return($ret) if ($ret);
	$ret = uc($a) cmp uc($b);
	return($ret) if ($ret);
	return(0);
}

sub byartist {
	my $ret = uc($a->{'artist'}) cmp uc($b->{'artist'});
	return($ret) if ($ret);
	return(0);
}

sub bysong {
	my $ret = uc($a->{'song'}) cmp uc($b->{'song'});
	return($ret) if ($ret);
	return(0);
}

sub byart {
	my $ret = uc($collection{$a}->{'artist'}) cmp uc($collection{$b}->{'artist'});
	return($ret) if ($ret);
	$ret = $a cmp $b;
	return($ret) if ($ret);
	return(0);
}




sub parsomatic {

	my ($file) = @_;
	my($albartist, $album, $artist, $total, $size, $totlength, $length, $qq, $song, $sort, $stuffptr, $track, %info, $year, $count);
	my @id = %playlist = %collection = ();

	open(FILE, "$file") or die "Could not open $file: $!\n"; 
	binmode(FILE); #for Non-Unix systems..
	

	#check the header
	if(getfoo(0, (length($ipodmagic)+2)/3) ne $ipodmagic)
	{
	die "err: could open $file, but:\nI don't think, thats an ipod - tunes db!\n";
	}

	$qq = 292; #the magic number!! (the HARDCODED start of the first mhit)
    $spacer = 156;
    if (getfoo($qq + $spacer, 4) ne "6d 68 6f 64") {
       $spacer = 244;
       if (getfoo($qq + $spacer, 4) ne "6d 68 6f 64") {
          die "err: Can't locate mhod w/ known spacers!\n";
       }
    }
# gnutunes header

  while($qq != -1) {
	%info = ();
	($qq,$stuffptr) = get_nod_a($qq,$spacer); #get_nod_a returns wher it's guessing the next MHIT, if it fails, it returns '-1'

	if (defined(%$stuffptr)) {
		%info = %$stuffptr;
	
		$artist = $info{'artist'};
		$file = $info{'file'};
		$year = (defined($info{'year'})) ? $info{'year'} : "";
		$sortart = (defined($info{'sortart'})) ? $info{'sortart'} : "";
		$song = $info{'song'};
		$album = ($info{'album'}) ? $info{'album'} : "$artist - no album";
		$track = (defined($info{'track'})) ? $info{'track'} : 0;
		$length = $info{'length'};
		$tr = 0;
		$disc = (defined($info{'disc'}) and $info{'disc'} > 0) ? $info{'disc'} : 1;
		
		$collection{$album}->{'sort'} = ($track) ? "track" : "artist" unless (defined($collection{$album}->{'sort'}));
		$sort = $collection{$album}->{'sort'};
		$tr = ($sort eq "artist") ? @{${$collection{$album}->{'tracks'}}[$disc-1]} : $track;
        	push(@dupes, "\n" . ${$collection{$album}->{'tracks'}}[$disc - 1]->[$tr]->{"artist"} . " - $album - tr. $tr - " . ${$collection{$album}->{'tracks'}}[$disc - 1]->[$tr]->{"song"},"$artist - $album - tr. $tr - $song") if (${$collection{$album}->{'tracks'}}[$disc - 1]->[$tr]->{"song"});
		${$collection{$album}->{'tracks'}}[$disc - 1]->[$tr]->{"song"} = $song;
		${$collection{$album}->{'tracks'}}[$disc - 1]->[$tr]->{"length"} = $length;
		${$collection{$album}->{'tracks'}}[$disc - 1]->[$tr]->{"file"} = $file;
		$collection{$album}->{'artist'} = $artist unless (defined($collection{$album}->{'artist'}));
		$collection{$album}->{'sortart'} = $sortart unless (defined($collection{$album}->{'sortart'}) and $collection{$album}->{'sortart'});
		$collection{$album}->{'year'} = $year unless (defined($collection{$album}->{'year'}));
		${$collection{$album}->{'tracks'}}[$disc - 1]->[$tr]->{"artist"} = $artist;
		$albartist = $collection{$album}->{'artist'};
		if ($artist ne $albartist) {
			$albartist = $album;
			$collection{$album}->{'artist'} = "$albartist:Various"
		}
		$total++;
		$totlength += $length;
		$size += $info{'size'};
		$id[$info{'id'}]->{'song'} = $song;
		$id[$info{'id'}]->{'artist'} = $artist;
		$id[$info{'id'}]->{'length'} = $length;
		print "." if ($total % 10 == 0 and lc($config{'ShowDots'}) eq 'yes' and $print);
	}
  }

  print "\n\nGetting playlists.\n" if ($print);
  $qq = getshoe(getshoe(4,4) + 8, 4) + 292;
  while($qq != -1) {
	($qq, $mpl, $cont, $plname) = get_pl($qq);
	next if ($mpl);
	print "." if ($qq != -1 and lc($config{'ShowDots'}) eq 'yes' and $print);
	$count = 0;
	foreach $song (@$cont) {
		$playlist{$plname}->[$count]->{'artist'} = $id[$song]->{'artist'};
		$playlist{$plname}->[$count]->{'song'} = $id[$song]->{'song'};
		$playlist{$plname}->[$count++]->{'length'} = $id[$song]->{'length'};
	}
  }

  print "\n" if ($print);

  close(FILE);
  return(\%collection,\%playlist,$total, $totlength, $size,\@dupes);
}


sub get_pl
{
my ($sum) = @_;
my($elx, $otxt, $len, $zip, $plname, $oid, $count, $mpl, @ret);
 if(getfoo($sum+20, 1) == 1)
 {
  return($sum + getshoe($sum+8,4),1);
 }

$sum += getshoe($sum+4,4);
while($len != -1) {
	$sum+=$len;
	($len, $mht, $px) = get_mhod($sum);
	$plname = $px if $mht == 1;
}


 while($zip !=-1)
 {
  $sum = $zip+$sum+76;
  ($zip, undef, $otxt) = get_mhod($sum);
  $elx = getshoe($sum-52, 4);                 #ugly hack, get a PL  item
   push(@ret,$elx) if $elx && getfoo($sum, 4) eq "6d 68 6f 64";
 }

if($plname)
{ return (($sum-76), $mpl, \@ret, $plname); }
else { return -1;}
}


sub get_nod_a {
my(@jerk, $sum, $spacer, $zip, $otxt, $oid);
my %stuff = ();
($sum,$spacer) = @_;
if(getfoo($sum, 4) eq "6d 68 69 74") #aren't we lost?
{

$stuff{'id'} = getshoe($sum+16 , 4);
$stuff{'size'}  = getshoe($sum+36 , 4);
$stuff{'length'} = int(getshoe($sum+40,4)/1000);
$stuff{'track'} = getshoe($sum+44,4); #song number
$stuff{'year'} = getshoe($sum+52,4); #year
$stuff{'disc'} = getshoe($sum+92,4);   #cd nr
$zip = 0;

   $sum += $spacer;
    while($zip != -1) {
     $sum = $zip+$sum; 
     ($zip, $oid, $otxt) = get_mhod($sum);    #returns the number where its guessing the next mhod, -1 if it's failed
      $jerk[$oid] = xmlstring($otxt);
    }

$stuff{'song'} = $jerk[1];
$stuff{'file'} = $jerk[2];
$stuff{'album'} = $jerk[3];
$stuff{'artist'} = $jerk[4];
$stuff{'sortart'} = $jerk[12];

#foreach $i (keys(%stuff)) {
#	print "$i - $stuff{$i}\n";
#}

#print ".";
return ($sum-$zip-1,\%stuff);          #black magic
}

  else {
  return "-1";
  }


}

#get a SINGLE mhod entry:
#
# get_mhod(START_OF_MHOD);
#
# return+seek = new_mhod should be there
sub get_mhod() {
my($seek, $xl, $ml, $mty, $foo, $id);
($seek) = @_;

$id = getfoo($seek, 4);                    #are we lost?

#print OUTHTML "ID: ".getstr($seek, 4)."\n";

  if($id ne "6d 68 6f 64") { $ml = -1;} #is the id INcorrect??
 
  else {
  #get the TYPE of the DB-Entry
$ml = getshoe($seek+8, 4);
$mty = getshoe($seek+12, 4);         #genre number
    if ($mty == 1 or $mty == 2 or $mty == 3 or $mty == 4 or $mty == 12) {
$xl = getshoe($seek+28,4);           #Entrylength

$foo = getstr($seek+40, $xl); #string of the entry
$foo =~ tr/\0//d; #we have many \0.. killem!
    }

  return ($ml, $mty, $foo);
 }
}


sub xmlstring
{
my($ret) = @_;
$ret =~ s/&/&amp;/g;
$ret =~ s/"/&quot;/g;
$ret =~ s/</&lt;/g;
$ret =~ s/>/&gt;/g;
$ret =~ s/'/&apos;/g;
return $ret;
}


sub uncode
{
my($ret) = @_;
$ret =~ s/&amp;/&/g;
$ret =~ s/&quot;/"/g;
$ret =~ s/&apos;/'/g;
return $ret;
}

sub getfoo {
#reads $anz chars from FILE and returns HEX values!
my($anz, $buffer, $xx, $xr, $start, $noseek);
($start, $anz, $noseek) = @_;
# paranoia checks
if(!$start) { $start = 0; }
if(!$anz) { $anz = "1"; }

#seek to the given position

seek(FILE, $start, 0);
#start reading
read(FILE, $buffer, $anz);
  foreach(split(//, $buffer)) {
$xx = sprintf("%02x ", ord($_));
$xr = "$xr$xx";
  }
 chop($xr);# no whitespace at end
 
 return $xr;
}

sub getshoe {
#reads $anz chars from FILE and returns int 
my($anz, $buffer, $xx, $xr, $start, $noseek, $xxt);
($start, $anz, $noseek) = @_;
# paranoia checks
if(!$start) { $start = 0; }
if(!$anz) { $anz = "1"; }

#seek to the given position

seek(FILE, $start, 0);
#start reading
read(FILE, $buffer, $anz);
   foreach(split(//, $buffer)) {
    $xx = sprintf("%02X", ord($_));
   $xr = "$xx$xr";
  }
  $xr = oct("0x".$xr);
 return $xr;
}







sub getstr {
#reads $anz chars from FILE and returns a string!
my($anz, $buffer, $xx, $xr, $start, $noseek);
($start, $anz, $noseek) = @_;
# paranoia checks
if(!$start) { $start = 0; }
if(!$anz) { $anz = "1"; }

#seek to the given position
#if 3th ARG isn't defined

seek(FILE, $start, 0);
#start reading
read(FILE, $buffer, $anz);

 return $buffer;
}

sub read_config_file {
	my ($file,$hash) = @_;
	my ($filebad, $key, $val);
	my %hash = %$hash;
	open(FI,$file) or $filebad = 1;
	if ($filebad) {
		print "\nCouldn't read $file.\n";
	} else {
		while (<FI>) {
			next if (/^#/);
			chomp;
			next if (/^\s*#/);
			s/^\s*//;
			s/\s*$//;
			($key, $value) = split(/[\s=>\-:]+/,$_,2);
			$hash{$key} = $value;
		}
	}
	return(\%hash);
}

sub read_command_line {
	my ($list,$hash) = @_;
	my %hash = %$hash;
	my @args = @$list;
	while (my $arg = shift(@args)) {
		if ($arg =~ /^-/) {
			$arg =~ s/^-//;
			my $val = shift(@args);
			$hash{$arg} = $val if ($val);
		}
	}
	return(\%hash);
}
